%header {
#include "grammar.hpp"
}

%source {
typedef std::vector<std::pair<std::string, bool>> vars_t;
typedef std::pair<std::vector<solar::symbol_t>, vars_t> rhs_t;
}


%type SYMBOL {std::string *} str
%destructor SYMBOL delete

%type SYMBOL_UC {std::string *} str
%destructor SYMBOL_UC delete

%type BLOCK {std::string *} str
%destructor BLOCK delete

%type SQBLOCK {std::string *} str
%destructor SQBLOCK delete

%type STRING {std::string *} str
%destructor STRING delete

%type CHAR {unsigned char} c


%type rhs {rhs_t *}
%destructor rhs delete

%type action {std::string *}
%destructor action delete

%type symbol {solar::symbol_t *}
%destructor symbol delete

%type term {solar::symbol_t *}
%destructor term delete

%type varname {std::string *}
%destructor varname delete


%extra_arg {__attribute__((unused)) solar::grammar_t *} grammar


grammar |=;
grammar |= grammar directive;


directive |= "%type" SYMBOL(nonterm) BLOCK(type) {
	grammar->nonterm_types.insert(std::make_pair(*nonterm, *type));
}

directive |= "%type" term(term) BLOCK(type) varname(name) {
	grammar->term_types.insert(std::make_pair(*term, std::make_pair(*type, *name)));
}

directive |= "%destructor" symbol(sym) varname(name) {
	grammar->destructors.insert(std::make_pair(*sym, *name));
}

directive |= "%source" BLOCK(block) {
	grammar->source_block = *block;
}

directive |= "%header" BLOCK(block) {
	grammar->header_block = *block;
}

directive |= "%extra_arg" BLOCK(type) varname(name) {
	grammar->extra_args.push_back(std::make_pair(*type, *name));
}

directive |= SYMBOL(lhs) "|=" rhs(rhs) action(action) {
	grammar->add_rule({solar::item_t(*lhs, rhs->first), rhs->second, *action});
}

directive |= SYMBOL(lhs) "|=" '(' CHAR(c1) "..." CHAR(c2) ')' ';' {
	vars_t vars;
	vars.emplace_back();

	for (unsigned int c = c1; c <= c2; c++)
		grammar->add_rule({solar::item_t(*lhs, {solar::symbol_t::make_char(c)}), vars, std::string()});
}


rhs |= [new rhs_t()]

rhs |= rhs(=rhs) symbol(sym) {
	rhs->first.push_back(*sym);
	rhs->second.emplace_back();

	return rhs;
}

rhs |= rhs(=rhs) symbol(sym) '(' varname(var) ')' {
	rhs->first.push_back(*sym);
	rhs->second.emplace_back(*var, true);

	return rhs;
}

rhs |= rhs(=rhs) symbol(sym) '(' '=' varname(var) ')' {
	rhs->first.push_back(*sym);
	rhs->second.emplace_back(*var, false);

	return rhs;
}

rhs |= rhs(=rhs) STRING(str) {
	for (char c : *str) {
		rhs->first.push_back(solar::symbol_t::make_char(c));
		rhs->second.emplace_back();
	}

	return rhs;
}


action |= ';' [new std::string]
action |= BLOCK(=v) [v]
action |= SQBLOCK(v) [new std::string("return " + *v + ";")]


symbol |= SYMBOL(v) [new solar::symbol_t(solar::symbol_t::make_nonterm(*v))]
symbol |= term(=v) [v]

term |= SYMBOL_UC(v) [new solar::symbol_t(solar::symbol_t::make_term(*v))]
term |= CHAR(v) [new solar::symbol_t(solar::symbol_t::make_char(v))]

varname |= SYMBOL_UC(=v) [v]
varname |= SYMBOL(=v) [v]
